#ifndef LIBVAL_H
#define LIBVAL_H

#include "UList"

class LibVal;

typedef UEList<LibVal>               LibValEList;
typedef LibValEList::iterator        LibValIter;

enum LibValType
{
  LIBERTY__VAL_STRING,
  LIBERTY__VAL_DOUBLE,
  LIBERTY__VAL_BOOLEAN,
  LIBERTY__VAL_INT,
  LIBERTY__VAL_UNDEFINED
};

class LibVal : public UEListNode
{
public:

  LibVal(LibValType t = LIBERTY__VAL_UNDEFINED) : d_type(t) {}
  ~LibVal(); 

  LibValType type() { return d_type; }
  void string_val(char* str) {
    uassert (d_type == LIBERTY__VAL_STRING);
    d_u.string_val = str;
  }
  void int_val(int i) {
    uassert (d_type == LIBERTY__VAL_INT);
    d_u.int_val = i;
  }
  void double_val(double d) {
    uassert (d_type == LIBERTY__VAL_DOUBLE);
    d_u.double_val = d;
  }
  void bool_val(bool b) {
    uassert (d_type == LIBERTY__VAL_BOOLEAN);
    d_u.bool_val = b;
  }
  
  const char* typeStr()    const;
  const char* string_val() const;
  int         int_val()    const;
  double      double_val() const;
  bool        bool_val()   const;

  void dump(FILE* f, const char* prepend, const char* append);

  LibVal* getThis() { return this; }

private:

  LibValType d_type;
  union
  {
    char* string_val;
    int int_val;
    double double_val;
    bool bool_val;
  } d_u;

};

#endif

