/******************************************************************************
    Copyright (c) 1996-2000 Synopsys, Inc.    ALL RIGHTS RESERVED

  The contents of this file are subject to the restrictions and limitations
  set forth in the SYNOPSYS Open Source License Version 1.0  (the "License"); 
  you may not use this file except in compliance with such restrictions 
  and limitations. You may obtain instructions on how to receive a copy of 
  the License at

  http://www.synopsys.com/partners/tapin/tapinprogram.html. 

  Software distributed by Original Contributor under the License is 
  distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either 
  expressed or implied. See the License for the specific language governing 
  rights and limitations under the License.

******************************************************************************/

#include <stdio.h>
#include "syntax.h"
#include <string.h>
#include <stdlib.h>
#include "mymalloc.h"

extern liberty_hash_table *libsynt_groups;
extern liberty_hash_table *libsynt_allgroups;
extern liberty_hash_table *libsynt_attrs;
#include <sys/types.h>

void dump_group(libsynt_group_info *g, FILE *outC, FILE *outH);
void dump_attr(libsynt_attribute_info *a, FILE *outC, FILE *outH);

#if HAVE_DIRENT_H
# include <dirent.h>
# define NAMLEN(dirent) strlen((dirent)->d_name)
#else
# define dirent direct
# define NAMLEN(dirent) (dirent)->d_namelen
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# if HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# if HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif

extern  libsynt_technology libsynt_techs[20];
extern  int libsynt_tech_count;
extern int syntax_parser_parse();

void syntax_parser_restart(FILE *fp);

static void
freeAttr(libsynt_attribute_info *attr)
{
  //printf("freeing %p (%s)\n", attr->name, attr->name);
  my_free((void*)attr->name);
  switch (attr->type) {
  case SYNTAX_ATTRTYPE_FLOAT:
    if (attr->u.floatcon) {
      my_free((void*)attr->u.floatcon->refd_default);
      my_free(attr->u.floatcon);
    }
    break;
  case SYNTAX_ATTRTYPE_INT:
    if (attr->u.intcon) {
      my_free(attr->u.intcon->refd_default);
      my_free(attr->u.intcon);
    }
    break;
  case SYNTAX_ATTRTYPE_ENUM:
    for (int i = attr->u.stringenum->size - 1; i >= 0; i--) {
      my_free((void*)attr->u.stringenum->array[i]);
    }
    my_free(attr->u.stringenum->array);
    my_free(attr->u.stringenum);
    break;
  case SYNTAX_ATTRTYPE_COMPLEX:
  case SYNTAX_ATTRTYPE_COMPLEX_UNKNOWN:
  case SYNTAX_ATTRTYPE_COMPLEX_LIST:
    { 
      libsynt_argument *arg, *arg_next;
      for (arg = attr->u.args; arg;) {
	arg_next = arg->next;
	my_free(arg);
	arg = arg_next;
      }
    }
    break;
  default:
    break;
  }
  my_free(attr);
}

static void
freeGroup(libsynt_group_info *group)
{
  libsynt_attribute_info *attr, *attr_next;
  libsynt_group_info *grp, *grp_next;

  my_free((void*)group->type);
  liberty_hash_destroy_hash_table (group->attr_hash);
  liberty_hash_destroy_hash_table (group->group_hash);
  for (attr = group->attr_list; attr;) {
    attr_next = attr->next;
    freeAttr(attr);
    attr = attr_next;
  }
  for (grp = group->group_list; grp;) {
    grp_next = grp->next;
    freeGroup (grp);
    grp = grp_next;
  }
  my_free(group);
}

static void
freeLib(libsynt_technology *lib)
{
  my_free((void*)lib->name);
  liberty_hash_destroy_hash_table(lib->group_hash);
  freeGroup(lib->lib);
}


static void
cleanup()
{
  liberty_hash_destroy_hash_table(libsynt_allgroups);
  liberty_hash_destroy_hash_table(libsynt_attrs);
  for (int i = 0; i < libsynt_tech_count; i++) {
    freeLib(libsynt_techs+i);
  }
  liberty_pool_free();
}

class UGlobalContainer;

UGlobalContainer *buildCBContainer() {
  return NULL;
}

void deleteCBContainer(UGlobalContainer* cont) {
}

void deleteAllStatics() {
}

int
main(int argc, char **argv)
{
  DIR *dirp;
  struct dirent *dp;
  char *p;
  FILE *fp, *outC, *outH;
  int i;
  liberty_hash_bucket *hb;
	
  liberty_pool_init ();

  libsynt_allgroups = liberty_hash_create_hash_table(503, 1, 0);
  libsynt_attrs = liberty_hash_create_hash_table(503, 1, 0);
	
  dirp = opendir("./desc");
  if (!dirp) {
    printf ("Error opening directory ./desc. Quitting!\n");
    cleanup();
    exit(0);			// exit ok to make "make run" pass!!
  }
  while( (dp = readdir(dirp)) != NULL ) {
    if( strncmp(dp->d_name,"syntax.",7) == 0 && strlen(dp->d_name) > 8) {
      char fname[1000];
      
      /* foreach file in the desc dir */
			
      /* form the master hash tables */
      libsynt_techs[libsynt_tech_count].group_hash = liberty_hash_create_hash_table(503, 1, 0);
      libsynt_groups = libsynt_techs[libsynt_tech_count].group_hash;
			
      libsynt_techs[libsynt_tech_count].name = (char*)my_malloc(NAMLEN(dp)+1);
      strcpy(libsynt_techs[libsynt_tech_count].name, dp->d_name+7);
      p = strchr(libsynt_techs[libsynt_tech_count].name,'.');
      if ( *p ) {
	*p = 0;
      }
      /* parse the file right here  -- the tech count gets bumped when
	 the file is parsed */
      sprintf(fname,"desc/%s",dp->d_name);

      printf("Processing %s\n", fname);
			
      fp = fopen(fname,"r");
      if (!fp) {
	printf("Had trouble opening %s. Fix the problem and re-run\n", fname);
	libsynt_tech_count++;
	cleanup();
	exit(23);
      }

      /* parse the file in question */
      try {
        syntax_parser_restart(fp);
        syntax_parser_parse();
      }
      catch (...)
      {
        // exception thrown by parsing... just exit.
        exit(0);
      }

      fclose(fp);
    }
  }
  closedir(dirp);
  libsynt_techs[libsynt_tech_count].name = 0;
	
  /* dump all of the 8 different structs defined in syntax.h, and the hash tables, too */

  printf("Generating syntax_decls.cc and syntax_decls.h...\n");
	
  outC = fopen("syntax_decls.cc","w");
  outH = fopen("syntax_decls.h","w");
  if ( !outC || !outH ) {
    printf("Couldn't open 'syntax_decls.cc' or 'syntax_decls.h'\n");
    printf("Solve the problem and re-run.\n");
    cleanup();
    exit(130);
  }

  fprintf(outC,"#include \"syntax.h\"\n");
  fprintf(outC,"#include \"syntax_decls.h\"\n");
	
  for (i=0;i< libsynt_tech_count; i++) {
    fprintf(outH,"\n\n/* Technology: %s       */\n\n",
	    libsynt_techs[i].name);
    fprintf(outC,"\n\n/* Technology: %s       */\n\n",
	    libsynt_techs[i].name);

    dump_group(libsynt_techs[i].lib, outC, outH);
    dump_group_hash(libsynt_techs[i].group_hash, outC, outH);
  }
  fprintf(outC,"libsynt_technology libsynt_techs[%d] = {\n", libsynt_tech_count );
  for (i=0;i< libsynt_tech_count; i++) {
    fprintf(outC,"{(char*)\"%s\", %s, %s },\n", 
	    libsynt_techs[i].name, make_rep("group", libsynt_techs[i].lib),
	    make_rep("ht", libsynt_techs[i].group_hash));
  }
  fprintf(outC,"};\n");
  fprintf(outC,"int libsynt_tech_num = %d;\n", libsynt_tech_count);
	
  fclose(outC);
  fclose(outH);

  printf("Generating attr_enum.h...\n");

  outH = fopen("attr_enum.h","w");

  if( !outH ) {
    printf("Couldn't open attr_enum.h. Fix the problem and rerun.\n");
    cleanup();
    exit(130);
  }
  fprintf(outH,"#ifndef ATTR_ENUM_H\n#define ATTR_ENUM_H\n");
  fprintf(outH,"#include <string.h>\n" );
  fprintf(outH,"\n\nenum attr_enum\n{\n" );

  i=1;
	
  fprintf(outH, "LIBERTY_ATTRENUM_invalid                            = 0,\n");
  for(hb = libsynt_attrs->all_list; hb ; hb=hb->all_next) {
    fprintf(outH,"LIBERTY_ATTRENUM_%s					= %d,\n", hb->name, i++);
  }
	
  fprintf(outH,"};\n\n#endif\n");
	
  fclose(outH);

  printf("Generating group_enum.h...\n");

  outH = fopen("group_enum.h","w");

  if( !outH ) {
    printf("Couldn't open group_enum.h. Fix the problem and rerun.\n");
    cleanup();
    exit(130);
  }
  fprintf(outH,"#ifndef GROUP_ENUM_H\n#define GROUP_ENUM_H\n");
  fprintf(outH,"#include <string.h>\n" );
  fprintf(outH,"\n\nenum group_enum\n{\n" );

  i=1;
	
  fprintf(outH, "LIBERTY_GROUPENUM_invalid                            = 0,\n");
  for(hb = libsynt_allgroups->all_list; hb ; hb=hb->all_next) {
    fprintf(outH,"LIBERTY_GROUPENUM_%s					= %d,\n", hb->name, i++);
  }
	
	
  fprintf(outH,"};\n\n#endif\n");
	
  fclose(outH);

  printf("Generating group_lookup.g...\n");

  outC = fopen("group_lookup.g","w");

  if( !outC ) {
    printf("Couldn't open group_lookup. Fix the problem and rerun.\n");
    cleanup();
    exit(130);
  }
	
  fprintf(outC,"%%{\n#include \"group_enum.h\"\n" );
  fprintf(outC,"#undef __GNUC__\n#undef __cplusplus\n%%}\n");
  fprintf(outC,"struct libGroupMap { const char *name; group_enum type; };\n%%%%\n");
	

  for(hb = libsynt_allgroups->all_list; hb ; hb=hb->all_next) {
    fprintf(outC,"%s,	LIBERTY_GROUPENUM_%s\n", hb->name,hb->name);
  }
	
  fprintf(outC,"%%%%\n");
	
  fclose(outC);

  printf("Generating attr_lookup.g...\n");

  outC = fopen("attr_lookup.g","w");

  if( !outC ) {
    printf("Couldn't open attr_lookup. Fix the problem and rerun.\n");
    cleanup();
    exit(130);
  }
	
  fprintf(outC,"%%{\n#include \"attr_enum.h\"\n");
  fprintf(outC,"#undef __GNUC__\n#undef __cplusplus\n%%}\n");
  fprintf(outC,"struct libAttrMap { const char *name; attr_enum type; };\n%%%%\n");
	

  for(hb = libsynt_attrs->all_list; hb ; hb=hb->all_next) {
    fprintf(outC,"%s,	LIBERTY_ATTRENUM_%s\n", hb->name,hb->name);
  }

  fprintf(outC,"%%%%\n");
	
  fclose(outC);

	
  printf("...Done.\n");
  cleanup();
  exit(0);
}

const char *attr_type_to_string(libsynt_attr_type t)
{
  switch( t ) {
  case SYNTAX_ATTRTYPE_STRING:
    return "SYNTAX_ATTRTYPE_STRING";
		
  case SYNTAX_ATTRTYPE_VIRTUAL:
    return "SYNTAX_ATTRTYPE_VIRTUAL";
		
  case SYNTAX_ATTRTYPE_ENUM:
    return "SYNTAX_ATTRTYPE_ENUM";
		
  case SYNTAX_ATTRTYPE_FLOAT:
    return "SYNTAX_ATTRTYPE_FLOAT";
		
  case SYNTAX_ATTRTYPE_INT:
    return "SYNTAX_ATTRTYPE_INT";
		
  case SYNTAX_ATTRTYPE_BOOLEAN:
    return "SYNTAX_ATTRTYPE_BOOLEAN";
		
  case SYNTAX_ATTRTYPE_COMPLEX:
    return "SYNTAX_ATTRTYPE_COMPLEX";

  case SYNTAX_ATTRTYPE_COMPLEX_UNKNOWN:
    return "SYNTAX_ATTRTYPE_COMPLEX_UNKNOWN";

  case SYNTAX_ATTRTYPE_COMPLEX_LIST:
    return "SYNTAX_ATTRTYPE_COMPLEX_LIST";
  }
  return NULL;  // to silence g++
}


/* dump attr */
void dump_attr(libsynt_attribute_info *a, FILE *outC, FILE *outH)
{
  int i;
  libsynt_argument *arg;

  if( !a )
    return;
	
  switch( a->type ) {
  case SYNTAX_ATTRTYPE_STRING:
    fprintf(outH,"extern libsynt_attribute_info attr_%lx;\n", (pint_t)a);
		
    fprintf(outC,"libsynt_attribute_info attr_%lx = { \"%s\", %s, %d, { 0 }, %s};\n",
	    (pint_t)a,
	    a->name,
	    attr_type_to_string(a->type), a->lineno, make_rep("attr",a->next));
				
    break;
		
  case SYNTAX_ATTRTYPE_VIRTUAL:
    fprintf(outH,"extern libsynt_attribute_info attr_%lx;\n", (pint_t)a);
		
    fprintf(outC,"libsynt_attribute_info attr_%lx = { \"%s\", %s, %d, { 0 }, %s};\n",
	    (pint_t)a,
	    a->name,
	    attr_type_to_string(a->type), a->lineno, make_rep("attr",a->next));
		
    break;
		
  case SYNTAX_ATTRTYPE_ENUM:
    fprintf(outC,"const char *enum_%lx[%d] = {",
	    (pint_t)a->u.stringenum->array, a->u.stringenum->size);
    for(i=0;i<a->u.stringenum->size;i++) {
      fprintf(outC,"\"%s\",", a->u.stringenum->array[i]);
    }
    fprintf(outC,"};\n");
    fprintf(outC,"libsynt_string_enum strenum_%lx = { enum_%lx, %d };\n", (pint_t)a->u.stringenum, (pint_t)a->u.stringenum->array, a->u.stringenum->size);
    fprintf(outH,"extern libsynt_attribute_info attr_%lx;\n", (pint_t)a);
		
    fprintf(outC,"libsynt_attribute_info attr_%lx = { \"%s\", %s, %d, { (libsynt_float_constraint *)(void *)%s }, %s};\n",
	    (pint_t)a,
	    a->name,
	    attr_type_to_string(a->type), a->lineno,
	    make_rep("strenum",a->u.stringenum), make_rep("attr",a->next));
		
    break;
		
  case SYNTAX_ATTRTYPE_FLOAT:
    if( a->u.floatcon && a->u.floatcon->refd_default )
      fprintf(outC,"libsynt_float_constraint floatcons_%lx = {%g, %g, %g, \"%s\", %d, %d, %d, %d};\n",
	      (pint_t)a->u.floatcon,
	      a->u.floatcon->greater_than, a->u.floatcon->less_than, a->u.floatcon->default_value,
	      a->u.floatcon->refd_default,
	      a->u.floatcon->greater_specd, a->u.floatcon->less_specd, a->u.floatcon->default_specd, a->u.floatcon->ref_default_specd );
		
    else if( a->u.floatcon && !a->u.floatcon->refd_default )
      fprintf(outC,"libsynt_float_constraint floatcons_%lx = {%g, %g, %g, 0, %d, %d, %d, %d};\n",
	      (pint_t)a->u.floatcon,
	      a->u.floatcon->greater_than, a->u.floatcon->less_than, a->u.floatcon->default_value,
	      a->u.floatcon->greater_specd, a->u.floatcon->less_specd, a->u.floatcon->default_specd, a->u.floatcon->ref_default_specd );
    fprintf(outH,"extern libsynt_attribute_info attr_%lx;\n", (pint_t)a);
    fprintf(outC,"libsynt_attribute_info attr_%lx = { \"%s\", %s, %d, { (libsynt_float_constraint *)(void *)%s }, %s};\n",
	    (pint_t)a,
	    a->name,
	    attr_type_to_string(a->type), a->lineno,
	    make_rep("floatcons",a->u.stringenum), make_rep("attr",a->next));
    break;
		
  case SYNTAX_ATTRTYPE_INT:
    if( a->u.intcon && a->u.intcon->refd_default )
      fprintf(outC,"libsynt_int_constraint intcons_%lx = {%d, %d, %d, \"%s\", %d, %d, %d, %d};\n",
	      (pint_t)a->u.intcon,
	      a->u.intcon->greater_than, a->u.intcon->less_than, a->u.intcon->default_value,
	      a->u.intcon->refd_default,
	      a->u.intcon->greater_specd, a->u.intcon->less_specd, a->u.intcon->default_specd, a->u.intcon->ref_default_specd );
		
    else if( a->u.intcon && !a->u.intcon->refd_default )
      fprintf(outC,"libsynt_int_constraint intcons_%lx = {%d, %d, %d, 0, %d, %d, %d, %d};\n",
	      (pint_t)a->u.intcon,
	      a->u.intcon->greater_than, a->u.intcon->less_than, a->u.intcon->default_value,
	      a->u.intcon->greater_specd, a->u.intcon->less_specd, a->u.intcon->default_specd, a->u.intcon->ref_default_specd );
    fprintf(outH,"extern libsynt_attribute_info attr_%lx;\n", (pint_t)a);
    fprintf(outC,"libsynt_attribute_info attr_%lx = { \"%s\", %s, %d, { (libsynt_float_constraint *)(void *)%s }, %s};\n",
	    (pint_t)a,
	    a->name,
	    attr_type_to_string(a->type), a->lineno,
	    make_rep("intcons",a->u.stringenum), make_rep("attr",a->next));
    break;
		
  case SYNTAX_ATTRTYPE_BOOLEAN:
    fprintf(outH,"extern libsynt_attribute_info attr_%lx;\n", (pint_t)a);
		
    fprintf(outC,"libsynt_attribute_info attr_%lx = { \"%s\", %s, %d, { 0 }, %s};\n",
	    (pint_t)a,
	    a->name,
	    attr_type_to_string(a->type), a->lineno, make_rep("attr",a->next));
		
    break;
		
  case SYNTAX_ATTRTYPE_COMPLEX:
  case SYNTAX_ATTRTYPE_COMPLEX_UNKNOWN:
  case SYNTAX_ATTRTYPE_COMPLEX_LIST:
    for(arg=a->u.args;arg;arg=arg->next) {
      fprintf(outH,"extern libsynt_argument arg_%lx;\n", (pint_t)arg);
      fprintf(outC,"libsynt_argument arg_%lx = {%s, %s};\n", (pint_t)arg, attr_type_to_string(arg->type), make_rep("arg",arg->next) );
    }
    fprintf(outH,"extern libsynt_attribute_info attr_%lx;\n", (pint_t)a);
    fprintf(outC,"libsynt_attribute_info attr_%lx = { \"%s\", %s, %d, { (libsynt_float_constraint *)(void *)%s }, %s};\n",
	    (pint_t)a,
	    a->name,
	    attr_type_to_string(a->type), a->lineno,
	    make_rep("arg",a->u.stringenum), make_rep("attr",a->next));
    break;
		
  }
}

const char *name_cons_to_str(libsynt_name_constraints x)
{
  switch( x ) {
  case SYNTAX_GNAME_NONE:
    return "SYNTAX_GNAME_NONE";
		
  case SYNTAX_GNAME_ONE:
    return "SYNTAX_GNAME_ONE";
		
  case SYNTAX_GNAME_TWO:
    return "SYNTAX_GNAME_TWO";
		
  case SYNTAX_GNAME_THREE:
    return "SYNTAX_GNAME_THREE";
		
  case SYNTAX_GNAME_NONE_OR_ONE:
    return "SYNTAX_GNAME_NONE_OR_ONE";
		
  case SYNTAX_GNAME_ONE_OR_MORE:
    return "SYNTAX_GNAME_ONE_OR_MORE";
  }
  return NULL;  // to silence g++
}

/* dump group */

void dump_group(libsynt_group_info *g, FILE *outC, FILE *outH)
{
  libsynt_group_info *ng;
  libsynt_attribute_info *at;

  g->mark = 1;
	
  for(ng = g->group_list; ng; ng=ng->next) {
    if( !ng->mark ) {
      dump_group(ng, outC, outH);
      ng->mark++;
    }
  }
  for(at = g->attr_list; at; at=at->next) {
    dump_attr(at, outC, outH);
  }
  dump_group_hash(g->group_hash, outC, outH);
  dump_attr_hash(g->attr_hash, outC, outH);
  fprintf(outC,"libsynt_group_info group_%lx = {%d, %s, \"%s\", %d, %d, %s, %s, %s, %s, %s, %s};\n",
	  (pint_t)g, g->mark, name_cons_to_str(g->name_constraints), g->type, g->ID, g->lineno,
	  make_rep("attr",g->attr_list), make_rep("group",g->group_list),
	  make_rep("ht",g->attr_hash), make_rep("ht",g->group_hash), make_rep("group",g->next), make_rep("group",g->ref));
  fprintf(outH,"extern libsynt_group_info group_%lx;\n", (pint_t)g);
}



